%{
	#include <stdlib.h>
	#include <stdio.h>
	#include <string>
	#include <hash_map.h>
	#include <stack.h>
	#include <list.h>
	
	using namespace std;
	
	namespace __gnu_cxx {
		template<> struct hash<std::string> {
			size_t operator()(const std::string& x) const {
				return hash<const char* >()(x.c_str());
			}
		};
	}
	
	extern FILE* yyin;
	extern int n;
	
	void przypisz(char*, int);
	int pobierz(char*);
	void inicjuj(char*);
	void nowy_kontekst();
	void stary_kontekst();
	
	void yyerror(char*);
	int yylex();
	
	stack<bool> warunki;
	list< hash_map<string, int> > konteksty;
	
	bool ie_war = true;
	bool ie_war_perm = true;
%}

%union {
	char* nazwa;
	int liczba;
}

%token print_t
%token srednik_t mul_t plus_t comp_t less_t more_t rown_t pnawias_t lnawias_t if_t else_t int_t pnawias_zw_t lnawias_zw_t
%token <liczba> stala_t
%token <nazwa> zmienna_t
%type <liczba> wartosc

%left less_t
%left more_t
%left comp_t
%left plus_t
%left mul_t

%%

program		:	instrukcja program
		|	instrukcja
;

instrukcja	:	print_t wartosc srednik_t { if(ie_war) printf("%d\n", $2); }
		|	zmienna_t rown_t wartosc srednik_t { if(ie_war) przypisz($1, $3); else free($1); }
		|	srednik_t
		|	lnawias_t { nowy_kontekst(); } program { stary_kontekst() } pnawias_t
		|	ifelse
		|	int_t zmienna_t srednik_t { inicjuj($2); }
;

ifelse		:	if_t lnawias_zw_t wartosc pnawias_zw_t {
				if (ie_war) {
					warunki.push(ie_war);
					if ($3)
						ie_war = true;
					else
						ie_war = false;
				} else
					ie_war_perm = false;
			} instrukcja else_t {
				if (ie_war_perm)
					ie_war = !ie_war;
			} instrukcja {
				if (ie_war_perm) {
					ie_war = warunki.top();
					warunki.pop();
				} else
					ie_war_perm = true;
			}
;

wartosc		:	stala_t { $$ = $1; }
		|	zmienna_t { $$ = pobierz($1); free($1); }
		|	wartosc mul_t wartosc { $$ = $1 * $3; }
		|	wartosc plus_t wartosc { $$ = $1 + $3; }
		|	wartosc comp_t wartosc { $$ = $1 == $3; }
		|	wartosc less_t wartosc { $$ = $1 < $3; }
		|	wartosc more_t wartosc { $$ = $1 > $3; }
;

%%

void nowy_kontekst() {
	hash_map<string, int> zmienne;
	konteksty.push_front(zmienne);
}

void stary_kontekst() {
	konteksty.pop_front();
}

void inicjuj(char* zm) {
	if (konteksty.begin()->find(zm) == konteksty.begin()->end())
		(*konteksty.begin())[zm]=0;
	else {
		fprintf(stderr, "Blad w linii %d: ponowna deklaracja zmiennej %s\n", n, zm);
		exit(EXIT_FAILURE);
	}
}

void przypisz(char* zm, int war) {
	list< hash_map<string, int> >::iterator it;
	
	for (it = konteksty.begin(); it != konteksty.end() && it->find(zm)==it->end(); ++it);
	
	if (it==konteksty.end()) {
		fprintf(stderr, "Blad w linii %d: niezadeklarowana zmienna %s\n", n, zm);
		exit(EXIT_FAILURE);
	}

	(*it)[zm] = war;
	free(zm);
}

int pobierz(char* zm) {
	list< hash_map<string, int> >::iterator it;
	
	for (it = konteksty.begin(); it != konteksty.end() && it->find(zm)==it->end(); ++it);
	
	if (it==konteksty.end()) {
		fprintf(stderr, "Blad w linii %d: niezadeklarowana zmienna %s\n", n, zm);
		exit(EXIT_FAILURE);
	}
	
	return (*it)[zm];
}

void yyerror(char*) {
	fprintf(stderr, "Blad w linii %d: nieoczekiwany token\n", n);
	exit(EXIT_FAILURE);
}

int main(int argc, char** argv) {
	if (argc<1) {
		fprintf(stderr, "Uzycie: %s plik\n", argv[0]);
		return EXIT_FAILURE;
	}

	yyin = fopen(argv[1], "r");

	if (!yyin) {
		fprintf(stderr, "Nie mozna otworzyc pliku\n");
		return EXIT_FAILURE;
	}
	
	hash_map<string, int> zmienne;
	konteksty.push_front(zmienne);
	
	yyparse();

	return EXIT_SUCCESS;
}
